#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <iostream>
#include <fstream>
#include <string>

#include "shaders/Shader.h"
#include "textures/Texture2D.h"
#include "util/stb_image.h"

// Continue: https://learnopengl.com/Getting-started/Camera
// Chapter: Walk around
//
// TODO: look at project->properties->VC++ Directories-> change everything to the Environment var
// For the "real" Nebulix Engine setup everything so it can be compiled/used with VSCode/CLion and not just VS 
// (so basically having an "external" folder which has open gl, GLAD, GLM, stb_image etc. in it or something like this)

// 
// HOW TO redirect error output to file
// Maybe it is possible to redirect to another stream object which adds infos like thread id/time etc.?
// 
// std::ofstream errorFile("error.log");
// std::streambuf* originalCerrBuffer = std::cerr.rdbuf();  // Save original buffer
// Redirect cerr to the error file
// std::cerr.rdbuf(errorFile.rdbuf());

#define WINDOW_WIDTH 800
#define WINDOW_HEIGHT 600
typedef glm::mat4 mat4;
typedef glm::vec3 vec3;


void OnWindowResize(GLFWwindow* window, int width, int height)
{
	glViewport(0, 0, width, height);
}

void ProcessInput(GLFWwindow* window)
{
	if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS)
		glfwSetWindowShouldClose(window, true);
}

void Configure()
{
	stbi_set_flip_vertically_on_load(true); // because an image has 0 at top of y axis and opengl expects it to be on the botton, so if this is omitted, the image will be flipped
	glEnable(GL_DEPTH_TEST);
}

int main()
{
	glfwInit();
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	GLFWwindow* window = glfwCreateWindow(WINDOW_WIDTH, WINDOW_HEIGHT, "Henlo worl", nullptr, nullptr);

	if (window == nullptr)
	{
		std::cerr << "ERROR::GLFW::CREATE_WINDOW";
		glfwTerminate();

		return -1;
	}

	glfwMakeContextCurrent(window);

	if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
	{
		std::cerr << "ERROR::GLAD::INITILISE";
		return -1;
	}

	OnWindowResize(window, WINDOW_WIDTH, WINDOW_HEIGHT);
	glfwSetFramebufferSizeCallback(window, OnWindowResize);

	Configure();

	float vertices[] = { // object coordinates, uv-coordinates
	-0.5f, -0.5f, -0.5f,  0.0f, 0.0f,
	 0.5f, -0.5f, -0.5f,  1.0f, 0.0f,
	 0.5f,  0.5f, -0.5f,  1.0f, 1.0f,
	 0.5f,  0.5f, -0.5f,  1.0f, 1.0f,
	-0.5f,  0.5f, -0.5f,  0.0f, 1.0f,
	-0.5f, -0.5f, -0.5f,  0.0f, 0.0f,

	-0.5f, -0.5f,  0.5f,  0.0f, 0.0f,
	 0.5f, -0.5f,  0.5f,  1.0f, 0.0f,
	 0.5f,  0.5f,  0.5f,  1.0f, 1.0f,
	 0.5f,  0.5f,  0.5f,  1.0f, 1.0f,
	-0.5f,  0.5f,  0.5f,  0.0f, 1.0f,
	-0.5f, -0.5f,  0.5f,  0.0f, 0.0f,

	-0.5f,  0.5f,  0.5f,  1.0f, 0.0f,
	-0.5f,  0.5f, -0.5f,  1.0f, 1.0f,
	-0.5f, -0.5f, -0.5f,  0.0f, 1.0f,
	-0.5f, -0.5f, -0.5f,  0.0f, 1.0f,
	-0.5f, -0.5f,  0.5f,  0.0f, 0.0f,
	-0.5f,  0.5f,  0.5f,  1.0f, 0.0f,

	 0.5f,  0.5f,  0.5f,  1.0f, 0.0f,
	 0.5f,  0.5f, -0.5f,  1.0f, 1.0f,
	 0.5f, -0.5f, -0.5f,  0.0f, 1.0f,
	 0.5f, -0.5f, -0.5f,  0.0f, 1.0f,
	 0.5f, -0.5f,  0.5f,  0.0f, 0.0f,
	 0.5f,  0.5f,  0.5f,  1.0f, 0.0f,

	-0.5f, -0.5f, -0.5f,  0.0f, 1.0f,
	 0.5f, -0.5f, -0.5f,  1.0f, 1.0f,
	 0.5f, -0.5f,  0.5f,  1.0f, 0.0f,
	 0.5f, -0.5f,  0.5f,  1.0f, 0.0f,
	-0.5f, -0.5f,  0.5f,  0.0f, 0.0f,
	-0.5f, -0.5f, -0.5f,  0.0f, 1.0f,

	-0.5f,  0.5f, -0.5f,  0.0f, 1.0f,
	 0.5f,  0.5f, -0.5f,  1.0f, 1.0f,
	 0.5f,  0.5f,  0.5f,  1.0f, 0.0f,
	 0.5f,  0.5f,  0.5f,  1.0f, 0.0f,
	-0.5f,  0.5f,  0.5f,  0.0f, 0.0f,
	-0.5f,  0.5f, -0.5f,  0.0f, 1.0f
	};

	GLuint faces[]{
		0, 1, 3,
		1, 2, 3
	};
	vec3 cubePositions[] = {
		vec3(0.0f,  0.0f,  0.0f),
		vec3(2.0f,  5.0f, -15.0f),
		vec3(-1.5f, -2.2f, -2.5f),
		vec3(-3.8f, -2.0f, -12.3f),
		vec3(2.4f, -0.4f, -3.5f),
		vec3(-1.7f,  3.0f, -7.5f),
		vec3(1.3f, -2.0f, -2.5f),
		vec3(1.5f,  2.0f, -2.5f),
		vec3(1.5f,  0.2f, -1.5f),
		vec3(-1.3f,  1.0f, -1.5f)
	};

	// VBO == all verticies; EBO == connection of vertices (faces) and is optional although it is a bit more efficient than using VBOs; 
	// VAO == collection of VBOs/EBOs used for rendering
	GLuint vertexBufferObject, elementBufferObject, vertexArrayObject;
	glGenVertexArrays(1, &vertexArrayObject);
	glGenBuffers(1, &vertexBufferObject);
	glGenBuffers(1, &elementBufferObject);

	glBindVertexArray(vertexArrayObject);

	glBindBuffer(GL_ARRAY_BUFFER, vertexBufferObject);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementBufferObject);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(faces), faces, GL_STATIC_DRAW);

	// position attribute
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
	glEnableVertexAttribArray(0);
	// uv-coordinates (or more general the third vertex attribute which can be anything not just uv-coordinates)
	glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
	glEnableVertexAttribArray(2);
	// I could unbind the VAO, but this is usually not needed because when creating a new VAO we need to bind that to change it anyway, so there shouldn't be a problem

	std::string vertexPath = "shaders/default/default.vert";
	std::string fragmentPath = "shaders/default/default.frag";

	std::unique_ptr<Nebulix::Shader> shader;
	try
	{
		shader = std::make_unique<Nebulix::Shader>(vertexPath, fragmentPath);
	}
	catch (const std::exception &e)
	{
		std::cerr << e.what();
		return -1;
	}
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
	// TEXTURES
	Texture2D containerImage("images/container.jpg");
	Texture2D faceImage("images/awesomeface.png", 0, GL_RGBA);


	vec3 cameraPosition = vec3(0.0f, 0, 3.0f);

	vec3 cameraTarget = vec3(0.0f, 0.0f, 0.0f);
	vec3 cameraDirection = glm::normalize(cameraPosition - cameraTarget);

	vec3 worldUp = vec3(0.0f, 1.0f, 0.0f);
	vec3 cameraRight = glm::normalize(glm::cross(worldUp, cameraDirection)); // vector pointing to the right of the camera

	vec3 cameraUp = glm::cross(cameraDirection, cameraRight);

	//mat4 viewMatrix = glm::lookAt(cameraPosition, cameraTarget, worldUp);
	mat4 projectionMatrix = glm::perspective(glm::radians(45.0f), 800.0f / 600.0f, 0.1f, 100.0f);

	float timeLastFrame = glfwGetTime();
	// main loop
	while(!glfwWindowShouldClose(window))
	{
		glClearColor(0.2f, 0.3f, 0.3f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		ProcessInput(window);

		const float radius = 10.0f;
		float camX = sin(glfwGetTime()) * radius;
		float camZ = cos(glfwGetTime()) * radius;
		mat4 viewMatrix = glm::lookAt(vec3(camX, 0, camZ), cameraTarget, worldUp);

		shader->Use();
		shader->SetInt("ourTexture1", 0);
		shader->SetInt("ourTexture2", 1);
		shader->SetMatrix("viewMatrix", viewMatrix); // more or less the camera
		shader->SetMatrix("projectionMatrix", projectionMatrix);

		containerImage.BindTexture();
		faceImage.BindTexture(GL_TEXTURE0 + 1); // GL_TEXTURE0 + 1 == GL_TEXTURE1, this means looping over multiple texture units is easily possible
		
		glBindVertexArray(vertexArrayObject);

		float time = glfwGetTime();
		for (int i = 0; i < 10; i++) // somehow cubePositions->length() does not return the correct number?
		{
			mat4 model = mat4(1.0f);
			model = glm::translate(model, cubePositions[i]);
			float angle = (20.0f * i);
			if (i % 3 == 0)
				angle = time * 25;
			model = glm::rotate(model, glm::radians(angle), glm::vec3(1.0f, 0.3f, 0.5f));
			shader->SetMatrix("modelMatrix", model);

			//glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0); // for EBOs
			glDrawArrays(GL_TRIANGLES, 0, 36); // for VAOs
		}

		glfwSwapBuffers(window);
		glfwPollEvents();
		timeLastFrame = glfwGetTime();
	}

	// optional, as we are at the end of the program.
	glDeleteVertexArrays(1, &vertexArrayObject);
	glDeleteBuffers(1, &elementBufferObject);

	glfwTerminate();
	return 0;
}